'use strict';
const {filter} = [];
const elements = element => 'querySelectorAll' in element;

module.exports = (options) => {
  const {cache, query, root} = options;
  const callback = records => {
    if (query.length) {
      for (let i = 0, {length} = records; i < length; i++) {
        const {addedNodes, removedNodes} = records[i];
        parse(filter.call(addedNodes, elements), true, new Set);
        parse(filter.call(removedNodes, elements), false, new Set);
      }
    }
  };
  const matches = element => (
    element.matches ||
    element.webkitMatchesSelector ||
    element.msMatchesSelector
  );
  const parse = (elements, connected, set) => {
    for (let element, i = 0, {length} = elements; i < length; i++) {
      if (!set.has(element = elements[i])) {
        set.add(element);
        if (cache.has(element))
          options.update(element, connected);
        else if (connected)
          setup(element);
        parse(element.querySelectorAll(query), connected, set);
      }
    }
  };
  const setup = element => {
    for (let m = matches(element), i = 0, {length} = query; i < length; i++) {
      if (m.call(element, query[i]))
        options.setup(element, query[i]);
    }
  };
  const observer = new MutationObserver(callback);
  observer.observe(root || document, {childList: true, subtree: true});
  if (query.length)
    parse((root || document).querySelectorAll(query), true, new Set);
  return {
    flush() {
      callback(observer.takeRecords());
    },
    matches,
    observer,
    parse(elements, connected = true) {
      parse(elements, connected, new Set);
    }
  };
};
